"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const stringHash = require("string-hash");
const Support_1 = require("./support/Support");
const Containers_1 = require("./support/Containers");
class TypeAttributeKind {
    constructor(name) {
        this.name = name;
    }
    combine(_a, _b) {
        return Support_1.panic(`Cannot combine type attribute ${this.name}`);
    }
    intersect(a, b) {
        return this.combine(a, b);
    }
    makeInferred(_) {
        return Support_1.panic(`Cannot make type attribute ${this.name} inferred`);
    }
    children(_) {
        return new Set();
    }
    stringify(_) {
        return undefined;
    }
    get inIdentity() {
        return false;
    }
    requiresUniqueIdentity(_) {
        return false;
    }
    reconstitute(_builder, a) {
        return a;
    }
    makeAttributes(value) {
        const kvps = [[this, value]];
        return new Map(kvps);
    }
    tryGetInAttributes(a) {
        return a.get(this);
    }
    setInAttributes(a, value) {
        // FIXME: This is potentially super slow
        return new Map(a).set(this, value);
    }
    modifyInAttributes(a, modify) {
        const modified = modify(this.tryGetInAttributes(a));
        if (modified === undefined) {
            // FIXME: This is potentially super slow
            const result = new Map(a);
            result.delete(this);
            return result;
        }
        return this.setInAttributes(a, modified);
    }
    combineInAttributes(a, value) {
        return this.modifyInAttributes(a, v => (v === undefined ? value : this.combine(v, value)));
    }
    setDefaultInAttributes(a, makeDefault) {
        if (this.tryGetInAttributes(a) !== undefined)
            return a;
        return this.modifyInAttributes(a, makeDefault);
    }
    removeInAttributes(a) {
        return Containers_1.mapFilter(a, (_, k) => k !== this);
    }
    equals(other) {
        if (!(other instanceof TypeAttributeKind)) {
            return false;
        }
        return this.name === other.name;
    }
    hashCode() {
        return stringHash(this.name);
    }
}
exports.TypeAttributeKind = TypeAttributeKind;
exports.emptyTypeAttributes = new Map();
function combineTypeAttributes(combinationKind, firstOrArray, second) {
    const union = combinationKind === "union";
    let attributeArray;
    let first;
    let rest;
    if (Array.isArray(firstOrArray)) {
        attributeArray = firstOrArray;
        if (attributeArray.length === 0)
            return exports.emptyTypeAttributes;
        first = new Map(attributeArray[0]);
        rest = attributeArray.slice(1);
    }
    else {
        if (second === undefined) {
            return Support_1.panic("Must have on array or two attributes");
        }
        first = new Map(firstOrArray);
        rest = [second];
    }
    for (const r of rest) {
        Containers_1.mapMergeWithInto(first, (aa, ab, kind) => (union ? kind.combine(aa, ab) : kind.intersect(aa, ab)), r);
    }
    return first;
}
exports.combineTypeAttributes = combineTypeAttributes;
function makeTypeAttributesInferred(attr) {
    return Containers_1.mapFilterMap(attr, (value, kind) => kind.makeInferred(value));
}
exports.makeTypeAttributesInferred = makeTypeAttributesInferred;
